/*
 * Copyright (c) 1995 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
#ifndef lint
static char copyright[] =
    "Copyright (c) 1995 Lawrence Berkeley Laboratory\nAll rights reserved.\n";
static char rcsid[] =
    "@(#) $Header: kgdbtunnel.c,v 1.3 96/01/15 00:09:22 leres Exp $ (LBL)";
#endif

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <netinet/in.h>

#include <arpa/inet.h>

#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include "gnuc.h"
#ifdef HAVE_OS_PROTO_H
#include "os-proto.h"
#endif

#include <err.h>


/* Globals */
static char kgdbname[] = "/dev/kgdb";
static char usage[] = "usage: %s [-dkpv] [-s speed] hostname[:port] [tty]";

static char *prog;
static int debug;
static int verbose;
static int speed = B38400;
#ifdef HAVE_OPENPTY
static int kgdbsymlink;
static int usepty;
#endif
static u_short port = 4040;
static int on = 1;
static char *device = kgdbname;

/* Externals */
extern int optind;
extern char *optarg;

/* Forwards */
static char *fmtdev(char *);
int main(int, char **);
#ifdef HAVE_OPENPTY
int openpty(int *, int *, char *, struct termios *, struct winsize *);
#endif
static int val2speed(int);
static void inittermios(struct termios *, int);
static void viswrite(const char *, char *, u_int);
static void dwrite(const char *, char *, u_int);
#ifdef HAVE_CFMAKERAW
void cfmakeraw(struct termios *);
#endif

int
main(argc, argv)
	int argc;
	char **argv;
{
	register int op, s, s2, i, cc, n;
	int fd, len;
	register char *cp, *hostname;
	register struct hostent *hp;
	struct in_addr addr;
	struct sockaddr_in skt_addr, rem_addr;
	struct termios termios;
	fd_set sfds, fds;
	char buf[1024];

	if ((cp = strrchr(argv[0], '/')) != NULL)
		prog = cp + 1;
	else
		prog = argv[0];

	while ((op = getopt(argc, argv, "dkpvs:")) != EOF)
		switch (op) {

		case 'd':
			++debug;
			break;

#ifdef HAVE_OPENPTY
		case 'k':
			++kgdbsymlink;
			++usepty;
			break;

		case 'p':
			++usepty;
			break;
#endif

		case 'v':
			++verbose;
			break;

		case 's':
			if ((i = atoi(optarg)) == 0 ||
			    (speed = val2speed(i)) == 0)
				err(1, usage, prog);
			break;

		default:
			fprintf(stderr, usage, prog);
			exit(1);
		}

	argc -= optind;
	argv += optind;

	/* One required argument (hostname) */
	if (argc < 1)
		err(1, usage, prog);

	/* Get ip address */
	hostname = strdup(argv[0]);
	if (hostname == NULL)
		err(1, "%s: strdup", prog);
	--argc;
	++argv;
	cp = strrchr(hostname, ':');
	if (cp != NULL) {
		*cp++ = '\0';
		port = atoi(cp);
		if (port == 0)
			err(1, usage, prog);
	}
	if (((int)(addr.s_addr = inet_addr(hostname))) == -1) {
		hp = gethostbyname(hostname);
		if (hp == NULL)
			err(1, "%s: unknown host %s", prog, hostname);
		bcopy((char *)hp->h_addr, (char *)&addr, sizeof(addr));
	}

	if (argc >= 1) {
		device = fmtdev(optarg);
		--argc;
		++argv;
	}
	if (argc != 0)
		err(1, usage, prog);

	/* Get tty and setup */
#ifdef HAVE_OPENPTY
	if (usepty) {
		/* XXX does speed matter with a pty? */
		char ptyname[128];
		int sfd;

		bzero(&termios, sizeof(termios));
		inittermios(&termios, speed);
		if (openpty(&fd, &sfd, ptyname, &termios, NULL) < 0)
			err(1, "openpty");
		if (chmod(ptyname, 0666) < 0)
			fprintf(stderr, "%s: chmod %s: %s",
			    prog, ptyname, strerror(errno));
		if (tcgetattr(fd, &termios) < 0)
			err(1, "tcgetattr");

		if (!kgdbsymlink)
			fprintf(stderr, "kgdbtunnel: pty is %s\n", ptyname);
		else {
			/* Remove old symlink */
			struct stat sbuf;

			if (lstat(kgdbname, &sbuf) < 0) {
				if (errno != ENOENT)
					err(1, "%s: lstat %s", prog, kgdbname);
			} else if (S_ISLNK(sbuf.st_mode)) {
				if (unlink(kgdbname) < 0)
					err(1, "%s: unlink %s", prog, kgdbname);
			}
			if (symlink(ptyname, kgdbname) < 0)
				err(1, "%s: symlink %s -> %s",
				    prog, kgdbname, ptyname);
			fprintf(stderr, "%s: pty is %s -> %s\n",
			    prog, kgdbname, ptyname);
		}
	} else
#endif
	{
		fd = open(device, O_RDWR);
		if (fd < 0) {
			fprintf(stderr, "%s: open: ", prog);
			perror(device);
			exit(1);
		}

		if (tcgetattr(fd, &termios) < 0)
			err(1, "tcgetattr");
		inittermios(&termios, speed);
		if (tcsetattr(fd, TCSAFLUSH, &termios) < 0)
			err(1, "tcsetattr");
		fprintf(stderr, "kgdbtunnel: tty is %s\n", device);
	}

	s = socket(AF_INET, SOCK_STREAM, 0);
	if (s < 0)
		err(1, "orig socket");
	if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR,
	    (char *)&on, sizeof(on)) < 0)
		err(1, "orig setsockopt");

	/* Try to originate first */
	bzero(&skt_addr, sizeof(skt_addr));
	skt_addr.sin_family = AF_INET;
	skt_addr.sin_addr.s_addr = INADDR_ANY;

	rem_addr = skt_addr;
	rem_addr.sin_family = AF_INET;
	rem_addr.sin_addr = addr;
	rem_addr.sin_port = htons(port);

	if (bind(s, (struct sockaddr *)&skt_addr, sizeof(skt_addr)) < 0)
		err(1, "%s: bind", prog);
	if (connect(s, (struct sockaddr *)&rem_addr, sizeof(rem_addr)) >= 0)
		fprintf(stderr, "%s: connected to %s.%d\n",
		    prog,
		    inet_ntoa(rem_addr.sin_addr),
		    (int)htons(rem_addr.sin_port));
	else if (errno != ECONNREFUSED)
		err(1, "%s: connect", prog);
	else {
		/* He's not listening so we will */
		(void)close(s);
		s = socket(AF_INET, SOCK_STREAM, 0);
		if (s < 0)
			err(1, "%s: wait socket", prog);
		if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR,
		    (char *)&on, sizeof(on)) < 0)
			err(1, "%s: wait setsockopt", prog);
		skt_addr.sin_port = htons(port);
		if (bind(s, (struct sockaddr *)&skt_addr, sizeof(skt_addr)) < 0)
			err(1, "%s: wait bind", prog);
		if (listen(s, 1) < 0)
			err(1, "%s: listen", prog);

		fprintf(stderr, "%s: waiting for %s.%d\n",
		    prog, hostname, (int)port);
		len = sizeof(rem_addr);
		s2 = accept(s, (struct sockaddr *)&rem_addr, &len);
		if (s2 < 0)
			err(1, "accept");
		close(s);
		s = s2;

		/* Make sure we got the right remote host */
		if (addr.s_addr != rem_addr.sin_addr.s_addr)
			err(1, "%s: accepted wrong host (0x%x != 0x%x)",
			    prog, addr.s_addr, rem_addr.sin_addr.s_addr);

		fprintf(stderr, "%s: connect from %s.%d\n", prog,
		    inet_ntoa(rem_addr.sin_addr),
		    (int)ntohs(rem_addr.sin_port));
	}

	/* Main loop */
	n = s;
	if (n < fd)
		n = s;
	++n;
	FD_ZERO(&sfds);
	FD_SET(s, &sfds);
	FD_SET(fd, &sfds);
	for (;;) {
		fds = sfds;
		i = select(n, &fds, NULL, NULL, NULL);
		if (i < 0)
			err(1, "%s: select", prog);
		if (i == 0) {
			fprintf(stderr, "%s: no fds ready!\n", prog);
			sleep(1);
			continue;
		}
		if (FD_ISSET(fd, &fds)) {
			errno = 0;
			cc = read(fd, buf, sizeof(buf));
			if (cc < 0)
				err(1, "%s: tty read", prog);
			if (cc == 0)
				err(1, "%s: tty EOF", prog);
			errno = 0;
			i = send(s, buf, cc, 0);
			if (i < 0)
				err(1, "%s: net send", prog);
			if (i != cc) {
				errno = 0;
				err(1, "%s: short net send", prog);
			}
			if (debug)
				dwrite("net>", buf, cc);
			if (verbose)
				viswrite("net>", buf, cc);
		}
		if (FD_ISSET(s, &fds)) {
			errno = 0;
			cc = recv(s, buf, sizeof(buf), 0);
			if (cc < 0)
				err(1, "%s: net recv", prog);
			if (cc == 0)
				err(1, "%s: net EOF", prog);
			errno = 0;
			i = write(fd, buf, cc);
			if (i < 0)
				err(1, "%s: tty write", prog);
			if (i != cc)
				err(1, "%s: short tty write", prog);
			if (debug)
				dwrite("tty>", buf, cc);
			if (verbose)
				viswrite("tty>", buf, cc);
		}
	}

	exit(0);
}

static void
inittermios(tp, speed)
	register struct termios *tp;
	register int speed;
{

	(void)cfsetispeed(tp, speed);
	(void)cfsetospeed(tp, speed);
#ifdef HAVE_CFMAKERAW
	cfmakeraw(tp);
#else
	tp->c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP|INLCR|IGNCR|ICRNL|IXON);
	tp->c_oflag &= ~OPOST;
	tp->c_lflag &= ~(ECHO|ECHONL|ICANON|ISIG|IEXTEN);
	tp->c_cflag &= ~(CSIZE|PARENB);
	tp->c_cflag |= CS8;
	tp->c_cc[VMIN] = 1;              /* ??? */
	tp->c_cc[VTIME] = 0;             /* ??? */
#endif
}

static char *
fmtdev(register char *cp)
{
	static char devbuf[64];

	if (cp[0] != '/') {
		sprintf(devbuf, "/dev/%.*s", (int)sizeof(devbuf) - 6, cp);
		cp = devbuf;
	}
	return (cp);
}

static struct speeds {
	int speed;		/* symbolic speed */
	int val;		/* numeric value */
} speeds[] = {
#ifdef B50
	{ B50,	50 },
#endif
#ifdef B75
	{ B75,	75 },
#endif
#ifdef B110
	{ B110,	110 },
#endif
#ifdef B134
	{ B134,	134 },
#endif
#ifdef B150
	{ B150,	150 },
#endif
#ifdef B200
	{ B200,	200 },
#endif
#ifdef B300
	{ B300,	300 },
#endif
#ifdef B600
	{ B600,	600 },
#endif
#ifdef B1200
	{ B1200, 1200 },
#endif
#ifdef B1800
	{ B1800, 1800 },
#endif
#ifdef B2400
	{ B2400, 2400 },
#endif
#ifdef B4800
	{ B4800, 4800 },
#endif
#ifdef B7200
	{ B7200, 7200 },
#endif
#ifdef B9600
	{ B9600, 9600 },
#endif
#ifdef B14400
	{ B14400, 14400 },
#endif
#ifdef B19200
	{ B19200, 19200 },
#endif
#ifdef B38400
	{ B38400, 38400 },
#endif
#ifdef B28800
	{ B28800, 28800 },
#endif
#ifdef B57600
	{ B57600, 57600 },
#endif
#ifdef B76800
	{ B76800, 76800 },
#endif
#ifdef B115200
	{ B115200, 115200 },
#endif
#ifdef B230400
	{ B230400, 230400 },
#endif
};
#define NSPEEDS (sizeof(speeds) / sizeof(speeds[0]))

static int
val2speed(val)
	register int val;
{
	register int n;
	register struct speeds *sp;

	for (sp = speeds, n = NSPEEDS; n > 0; ++sp, --n)
		if (val == sp->val)
			return (sp->speed);

	return (0);
}

static void
viswrite(prefix, str, n)
	register const char *prefix;
	register char *str;
	register u_int n;
{
	register char c;

	fprintf(stderr, "%s: \"", prefix);
	while (n-- > 0) {
		c = *str++;
		if (c & ~ 0177) {
			(void)fputs("M-", stderr);
			c &= 0177;
		}
		if (c >= 0x20 && c != 0177) {
			(void)fputc(c, stderr);
			continue;
		}
		if (c == '\0')
			(void)fputs("\\0", stderr);
		else if (c == '\b')
			(void)fputs("\\b", stderr);
		else if (c == '\f')
			(void)fputs("\\f", stderr);
		else if (c == '\n')
			(void)fputs("\\n", stderr);
		else if (c == '\r')
			(void)fputs("\\r", stderr);
		else if (c == '\t')
			(void)fputs("\\t", stderr);
		else if (c == '"')
			(void)fputs("\\\"", stderr);
		else {
			(void)fputc('^', stderr);
			/* DEL to ?, others to alpha */
			(void)fputc(c ^ 0x40, stderr);
		}
	}
	fprintf(stderr, "\"\n");
}

static void
dwrite(prefix, str, n)
	register const char *prefix;
	register char *str;
	register u_int n;
{
	register char c;

	fprintf(stderr, "%s:", prefix);
	while (n-- > 0) {
		c = *str++;
		fprintf(stderr, " 0x%02x", c & 0xff);
	}
	fprintf(stderr, "\n");
}
